/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikaoverlaychild.h
 * Copyright (C) 2009 Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __PIKA_OVERLAY_CHILD_H__
#define __PIKA_OVERLAY_CHILD_H__


typedef struct _PikaOverlayChild PikaOverlayChild;

struct _PikaOverlayChild
{
  GtkWidget      *widget;
  GdkWindow      *window;

  gboolean        has_position;
  gdouble         xalign;
  gdouble         yalign;
  gdouble         x;
  gdouble         y;

  gdouble         angle;
  gdouble         opacity;

  /* updated in size_allocate */
  cairo_matrix_t  matrix;
};


PikaOverlayChild * pika_overlay_child_new                  (PikaOverlayBox  *box,
                                                            GtkWidget       *widget,
                                                            gdouble          xalign,
                                                            gdouble          yalign,
                                                            gdouble          angle,
                                                            gdouble          opacity);
void               pika_overlay_child_free                 (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child);

PikaOverlayChild * pika_overlay_child_find                 (PikaOverlayBox   *box,
                                                            GtkWidget        *widget);

void               pika_overlay_child_realize              (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child);
void               pika_overlay_child_unrealize            (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child);
void               pika_overlay_child_get_preferred_width  (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child,
                                                            gint             *minimum,
                                                            gint             *natural);
void               pika_overlay_child_get_preferred_height (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child,
                                                            gint             *minimum,
                                                            gint             *natural);
void               pika_overlay_child_size_allocate        (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child);
gboolean           pika_overlay_child_draw                 (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child,
                                                            cairo_t          *cr);
gboolean           pika_overlay_child_damage               (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child,
                                                            GdkEventExpose   *event);

void               pika_overlay_child_invalidate           (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child);
gboolean           pika_overlay_child_pick                 (PikaOverlayBox   *box,
                                                            PikaOverlayChild *child,
                                                            gdouble           box_x,
                                                            gdouble           box_y);


#endif /* __PIKA_OVERLAY_CHILD_H__ */
